package soot.javaToJimple.ppa;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import polyglot.ast.Node;
import polyglot.ext.jl.ast.SourceFile_c;
import polyglot.frontend.Job;
import polyglot.frontend.Pass;
import polyglot.frontend.VisitorPass;
import polyglot.visit.NodeVisitor;
import soot.javaToJimple.CastInsertionVisitor;
import soot.javaToJimple.JavaToJimple;
import soot.javaToJimple.SaveASTVisitor;
import soot.javaToJimple.ppa.jj.PPAExtensionInfo;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem_c;

public class PPAEngine {

	public static int NO_PPA = 0;

	public static int NO_INFERENCE_MODE = 1;

	public static int SINGLE_MODE = 2;

	public static int COLLECTIVE_MODE = 3;

	public static int DEFAULT_MODE = COLLECTIVE_MODE;

	public static int MAX_TURN = 10000000;

	public static final polyglot.frontend.Pass.ID CAST_INSERTION = JavaToJimple.CAST_INSERTION;

	public static polyglot.frontend.Pass.ID SAVE_AST = JavaToJimple.SAVE_AST;

	// static initialization must come after all other static initializations...
	private static PPAEngine instance = new PPAEngine();

	private static PPATypeSystem typeSystem;

	private boolean forceBinding = true;

	public static PPAEngine v() {
		return instance;
	}

	private int mode;

	// The worklist contains all nodes to be processed because new types facts were inferred.
	private final List<TypeFact> worklist;

	private final PPATypeMerger merger;

	private final Map<PPAIndex, List<UnsafeTuple>> unsafeTuples;

	private final List<SourceFile_c> parsedFiles;

	private boolean processing = false;

	private int turn = 0;
	
	private long turn2 = 0;

	private TypeFact currentTuple;

	private PPAExtensionInfo info = null;

	private final Set<String> acceptedPaths;

	private Map<String, Set<String>> classFileIndex = null;

	public double turnRatio = 0.0;
	
	public int maxNodes = 0;
	
	public long totalTurns = 0;

	public long totalNodes = 0;
	
	private PPAEngine() {
		this.mode = DEFAULT_MODE;
		typeSystem = new PPATypeSystem_c();
		worklist = new LinkedList<TypeFact>();
		parsedFiles = new ArrayList<SourceFile_c>();
		unsafeTuples = new HashMap<PPAIndex, List<UnsafeTuple>>();
		merger = new PPATypeMerger();
		acceptedPaths = new HashSet<String>();
	}

	public Map<String, Set<String>> getClassFileIndex() {
		return classFileIndex;
	}

	public void setClassFileIndex(Map<String, Set<String>> classFileIndex) {
		this.classFileIndex = classFileIndex;
	}

	public String getFile(String className) {
		String index = null;

		for (String key : classFileIndex.keySet()) {
			if (classFileIndex.get(key).contains(className)) {
				index = key;
				break;
			}
		}

		return index;
	}

	public boolean hasFileIndex() {
		return classFileIndex != null;
	}

	/**
	 * <p>
	 * Add a source file that will be processed by PPAEngine.
	 * </p>
	 * 
	 * @param node
	 */
	public void addSourceFile(SourceFile_c node) {
		String path = node.source().path();
		boolean canAdd = true;

		for (SourceFile_c temp : parsedFiles) {
			if (temp.source().path().equals(path)) {
				canAdd = false;
				break;
			}
		}

		if (canAdd) {
			parsedFiles.add(node);
		}
	}

	public PPAExtensionInfo getExtensionInfo() {
		// if (info == null) {
		// info = new PPAExtensionInfo();
		// }

		return info;
	}

	public PPATypeSystem getTypeSystem() {
		return typeSystem;
	}

	public Set<String> getAcceptedPaths() {
		return acceptedPaths;
	}

	public int getMode() {
		return mode;
	}

	public boolean isForceBinding() {
		return forceBinding;
	}

	public void setForceBinding(boolean forceBinding) {
		this.forceBinding = forceBinding;
	}

	public polyglot.frontend.ExtensionInfo initExtInfo(String fileName, List sourceLocations) {

		Set source = new HashSet();
		boolean needInit = false;
		if (info == null) {
			needInit = true;
			info = new PPAExtensionInfo(getTypeSystem()) {
				@Override
				public List passes(Job job) {
					List passes = super.passes(job);
					// beforePass(passes, Pass.EXIT_CHECK, new
					// VisitorPass(polyglot.frontend.Pass.FOLD,
					// job, new polyglot.visit.ConstantFolder(ts, nf)));
					// beforePass(passes, Pass.EXIT_CHECK, new VisitorPass(CAST_INSERTION, job, new
					// CastInsertionVisitor(job, ts, nf)));
					// beforePass(passes, Pass.EXIT_CHECK, new VisitorPass(STRICTFP_PROP, job, new
					// StrictFPPropagator(false)));
					// beforePass(passes, Pass.EXIT_CHECK, new VisitorPass(ANON_CONSTR_FINDER, job,
					// new
					// AnonConstructorFinder(job, ts, nf)));
					beforePass(passes, Pass.EXIT_CHECK, new VisitorPass(CAST_INSERTION, job,
							new CastInsertionVisitor(job, ts, nf)));
					afterPass(passes, Pass.PRE_OUTPUT_ALL, new SaveASTVisitor(SAVE_AST, job, this));
					removePass(passes, Pass.OUTPUT);
					return passes;
				}
			};
		}
		polyglot.main.Options options = info.getOptions();

		if (needInit) {
			options.assertions = true;
			options.source_path = new LinkedList();
			Iterator it = sourceLocations.iterator();
			while (it.hasNext()) {
				Object next = it.next();
				// System.out.println("adding src loc: "+next.toString());
				options.source_path.add(new File(next.toString()));
			}

			options.source_ext = new String[] { "java" };
			options.serialize_type_info = false;
		}

		if (!source.contains(fileName)) {
			source.add(fileName);
		}

		File parentFile = new File(fileName).getParentFile();
		if (!options.source_path.contains(parentFile)) {
			options.source_path.add(parentFile);
		}

		polyglot.main.Options.global = options;

		return info;
	}

	/**
	 * 
	 * @param tuple
	 * @return True if the given tuple is the same as the tuple being currently processed by the ppa
	 *         engine.
	 */
	private boolean isDuplicate(TypeFact tuple) {
		return processing && currentTuple != null
				&& currentTuple.getIndex().equals(tuple.getIndex());
	}

	private void putTuple(UnsafeTuple tuple) {
		PPAIndex key = tuple.getIndex();
		List<UnsafeTuple> list = unsafeTuples.get(key);
		if (list == null) {
			list = new ArrayList<UnsafeTuple>();
			unsafeTuples.put(key, list);
		}

		try {
			if (!list.contains(tuple)) {
				list.add(tuple);
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
	}

	/**
	 * <p>
	 * Report that a node has been made safer, i.e., one type fact associated with the node is more
	 * precise.
	 * </p>
	 * <p>
	 * e.g., secretMethod(MagicClass) is now secretMethod(String)
	 * </p>
	 * 
	 * @param tuple
	 */
	public void reportNewFact(TypeFact tuple) {

		// If
		if (isDuplicate(tuple)) {
			// reportError("This tuple is a duplicate " + tuple.getIndex() + " for turn: " + turn);
			return;
		}

		// if (tuple.toString().contains("declaringClass") && tuple.toString().contains("char[]")) {
		// System.out.println();
		// }

		PPAIndex index = tuple.getIndex();
		TypeFact newTuple = tuple;

		if (index == null) {
			return;
		}

		if (!TypeFactUtil.hasNewTypeInfo(newTuple)) {
			return;
		}

		if (!TypeFactUtil.isTypeSafer(newTuple.getNewType(), newTuple.getOldType())) {
			return;
		}

		for (TypeFact temp : worklist) {
			// Found two safer op on the same.
			if (temp.getIndex().equals(index)) {
				newTuple = merger.merge(temp, tuple);
				worklist.remove(temp);
				break;
			}
		}

		worklist.add(newTuple);
	}

	/**
	 * <p>
	 * Report that an ast node is unsafe, i.e., it contains an imprecise type fact.
	 * </p>
	 * <p>
	 * e.g., secretMethod(MagicClass) contains an imprecise type fact (the first argument).
	 * </p>
	 * 
	 * @param index
	 * @param unsafeNode
	 */
	public void reportUnsafe(PPAIndex index, PPANode unsafeNode) {
		UnsafeTuple tuple = new UnsafeTuple(index, unsafeNode);
		putTuple(tuple);
	}

	public void reset() {
		worklist.clear();
		parsedFiles.clear();
		unsafeTuples.clear();
		typeSystem = new PPATypeSystem_c();
		info = null;
		turn = 0;
		processing = false;
		currentTuple = null;
		acceptedPaths.clear();
		mode = DEFAULT_MODE;
		// PPAReport.v().reset();
	}

	public void setMode(int mode) {
		this.mode = mode;
	}

	/**
	 * Do not reset accepted paths.
	 */
	public void softReset() {
		worklist.clear();
		parsedFiles.clear();
		unsafeTuples.clear();
		typeSystem = new PPATypeSystem_c();
		info = null;
		turn = 0;
		processing = false;
		currentTuple = null;
		PPAReport.v().softReset();
		// mode = DEFAULT_MODE;
	}

	/**
	 * <p>
	 * Start PPA inference on all source files processed by PPA.
	 * </p>
	 * 
	 * <p>
	 * The inference goes as follow:
	 * <ol>
	 * <li>UnsafeNodeVisitor marks ast nodes that are unsafe, i.e., that contain imprecise type
	 * fact (e.g., reference to MagicClass)</li>
	 * <li>A first inference pass is done to empty the worklist (which is dynamic in size during
	 * this process).</li>
	 * <li>UnsafeCallVisitor marks call nodes that are still unsafe, i.e., that can be bound to
	 * more than one method.</li>
	 * <li>A second inference and binding pass is done to empty the worklist (which is dynamic in
	 * size during this process).</li>
	 * </ol>
	 * </p>
	 */
	public void makeNodesSafer() {
		if (mode == NO_INFERENCE_MODE) {
			return;
		}

		NodeVisitor visitor = null;
		// Index unsafe nodes!
		visitor = new UnsafeNodeVisitor();
		for (Node node : parsedFiles) {
			node.visit(visitor);
		}
		processWorklist();

		if (!forceBinding) {
			return;
		}

		visitor = new UnsafeCallVisitor();
		for (Node node : parsedFiles) {
			node.visit(visitor);
		}
		processWorklist();
	}

	/**
	 * <p>
	 * For each type fact in the worklist, get all unsafe nodes related to this type fact. First,
	 * process all such nodes which have a primary index equals to the new type fact index. Second,
	 * process all such nodes which have a secondary index equals to the new type fact index. We
	 * then recompute the indexes of nodes that were changed.
	 * </p>
	 * 
	 * <p>
	 * 
	 * </p>
	 */
	private void processWorklist() {
		turn = 0;
		turn2 = 0;
		int workListLength = worklist.size();
		int impactedNodes = 0;
		processing = true;

		try {

			while (!worklist.isEmpty()) {
				turn++;
				impactedNodes = 0;
				TypeFact currentTypeFact = worklist.get(0);

				if (turn > MAX_TURN) {
					System.err.println("INIFINITE LOOP!");
					break;
				}

				if (turn % 100 == 0) {
					System.out.println(turn + " : " + worklist.size());
				}
				// //
				// if (worklist.size() == 1) {
				// System.out.println("turn!");
				// }
				//				
				currentTuple = currentTypeFact;
				worklist.remove(0);
				List<UnsafeTuple> tuplesToRecompute = new ArrayList<UnsafeTuple>();
				List<UnsafeTuple> visitInSecond = new ArrayList<UnsafeTuple>();
				PPAIndex index = currentTypeFact.getIndex();

				// First visit nodes where mainIndex = index
				if (unsafeTuples.containsKey(index)) {
					if (unsafeTuples.get(index).size() > 0) {
						turn2++;
					}
					for (UnsafeTuple tuple : unsafeTuples.get(index)) {
						impactedNodes++;
						PPANode unsafeNode = tuple.getUnsafeNode();
						PPAIndex mainIndex = unsafeNode.getMainIndex();
						if (mainIndex == null || !mainIndex.equals(index)) {
							visitInSecond.add(tuple);
						} else {
							unsafeNode.makeSafe(currentTypeFact.getIndex(), currentTypeFact);
							tuplesToRecompute.add(tuple);
						}
					}
				} else {
					// System.err.println("Interesting");
				}

				// Then visit other unsafe nodes
				for (UnsafeTuple tuple : visitInSecond) {
					tuple.getUnsafeNode().makeSafe(currentTypeFact.getIndex(), currentTypeFact);
					tuplesToRecompute.add(tuple);
				}

				// XXX We do not remove indexes, we just add new ones.
				// This is because We would then need to change the index of nodes
				// in worklist pointing to them...
				// TODO Not removing indexes: not necessarily... Will need to check that!!
				recomputeIndexes(tuplesToRecompute);

				PPAReport.v().reportTypeFact(currentTypeFact);
				
				if (impactedNodes > maxNodes) {
					maxNodes = impactedNodes;
				}
				totalNodes += impactedNodes;
				

			}
		} catch (RuntimeException e) {
			System.err.println("Error occurred while processing worklist : " + e.getMessage());
			throw e;
		} finally {
			// System.err.println("Performed " + turn + " turns.");
			double ratio = (double) turn / (double) workListLength;
			if (ratio > turnRatio) {
				turnRatio = ratio;
			}
			turn = 0;
			totalTurns += turn2;
			processing = false;
			currentTuple = null;
		}
	}

	/**
	 * <p>
	 * We basically retrieve all indexes of nodes that were recomputed and put them in the
	 * unsafeTuples index.
	 * </p>
	 * 
	 * @param tuplesToRecompute
	 */
	private void recomputeIndexes(List<UnsafeTuple> tuplesToRecompute) {
		for (UnsafeTuple toRecompute : tuplesToRecompute) {
			PPANode node = toRecompute.getUnsafeNode();
			for (PPAIndex newIndex : node.getIndexes()) {
				putTuple(new UnsafeTuple(newIndex, node));
			}
		}
	}

}
