package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Field;
import polyglot.ast.Node;
import polyglot.ext.jl.ast.FieldAssign_c;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.AscriptionVisitor;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.ppa.PPAAssignHelper;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;

/**
 * <p>Field assignment. E.g. myObj.field1 = 2;</p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li>Based on PPAAssignHelper.</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>the node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>left node index</li>
 * <li>right node index</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Left and Right nodes are safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 *
 */
public class PPAFieldAssign_c extends FieldAssign_c implements PPANode {

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		
		indexes.add(getMainIndex());
		
		if (!TypeFactUtil.isSafe(left)) {
			indexes.add(TypeFactUtil.getMainIndex(left));
		}

		if (!TypeFactUtil.isSafe(right)) {
			indexes.add(TypeFactUtil.getMainIndex(right));
		}

		return indexes;
	}

	// Works because this node should never be promoted in the worklist (indexed by main index).
	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(left) && TypeFactUtil.isSafe(right);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		try {
			inferTypeInfo(null, newType, origin);
			type = newType;
		} catch (Exception e) {
		}

		// return isSafe();
	}

	public PPAFieldAssign_c(Position pos, Field left, Operator op, Expr right) {
		super(pos, left, op, right);
	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;
		SemanticException se = null;
		try {
			node = super.typeCheck(tc);
		} catch (SemanticException seTemp) {
			se = seTemp;
		}
		inferTypeInfo(se, null, null);
		node = type(left.type());

		return node;
	}

	private void inferTypeInfo(SemanticException se, Type newType, TypeFact origin) throws SemanticException {
		PPAAssignHelper.inferTypeInfo(left, right, op, se, newType, origin);
	}

	public Type childExpectedType(Expr child, AscriptionVisitor av) {
		if (op == SHL_ASSIGN || op == SHR_ASSIGN || op == USHR_ASSIGN) {
			return child.type();
		}

		if (child == right) {
			return left.type();
		}

		return child.type();

	}
	
	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPAFieldAssign_c) {
			PPAFieldAssign_c ppaObj = (PPAFieldAssign_c) obj;
			return TypeFactUtil.positionEquals(position, ppaObj.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}
}
