package soot.javaToJimple.ppa.jj.types;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import polyglot.frontend.Compiler;
import polyglot.frontend.ExtensionInfo;
import polyglot.frontend.FileSource;
import polyglot.types.Named;
import polyglot.types.NoClassException;
import polyglot.types.SemanticException;
import polyglot.types.SourceClassResolver;
import polyglot.types.TypeSystem;
import polyglot.types.reflect.ClassFileLoader;
import soot.javaToJimple.ppa.PPAEngine;
import soot.options.Options;

/**
 * <p>
 * This class is responsible for creating missing types.
 * </p>
 *
 * <p>
 * Basically, it records every request made to the resolver and when the request is , it
 * creates the most plausible type.
 * <p>
 *
 * <p>
 * A request is  if return an exception or a null type would result in a polyglot compilation
 * error.
 * </p>
 *
 * <p>
 * The rules for generating missing types are:
 * <ol>
 * <li>If there was only one request (e.g., foo.bar.MissingType), this type is returned.</li>
 * <li>If there were multiple requests, but on the same package family, the longest package is used
 * (e.g., MissingType, foo, foo.bar, foo.bar.MissingType: the last one will be returned.).</li>
 * <li>If there were multiple requests from different package families, we return the short name
 * with the magic package.</li>
 * </ol>
 * </p>
 *
 * @author barthelemy
 *
 */
public class PPASourceClassResolver extends SourceClassResolver {

	private final  TypeSystem ts;

	// Contains all name resolution attempts.
	private final  List<String> names;

	public PPASourceClassResolver(Compiler compiler, ExtensionInfo ext, String classpath,
			ClassFileLoader loader, boolean allowRawClasses, TypeSystem ts) {
		super(compiler, ext, classpath, loader, allowRawClasses);
		this.ts = ts;
		names = new ArrayList<String>();
	}

	@Override
	public Named find(String name, boolean requestFinal) throws SemanticException {
		Named named = null;

		// XXX THIS IS A HACK! Because Compiler is used, but obviously, not the one provided by
		// java.lang... Duh!
//		if (name.equals("java.lang.Compiler") || name.equals("java.text.ParseException")) {
//			throw new NoClassException(name);
//		}

		try {
			named = super.find(name, requestFinal);
			names.clear();
		} catch ( SemanticException e) {
			if (requestFinal) {
				 String fullName = getFullName(name);

//				System.err.println("Will return a magic class: " + fullName);
				
//				PPAExtensionInfo info = PPAEngine.v().getExtensionInfo();
//				System.err.println("In: " + info.getCurrentJob().toString() + " : " + info.getCurrentJob().runningPass().toString());
				
				 MagicClass mClass = new MagicClass(getShortName(fullName),
						getPackageName(fullName), fullName, ts);
				named = mClass;
			} else {
				names.add(name);
				throw e;
			}
		}

		return named;
	}

	private String getFullName(String lastName) {
		String fullName = lastName;
		 String shortName = getShortName(lastName);

		// Only search for better names if the last name is not a FQN.
		if (fullName.indexOf(".") == -1) {
			int found = 0;

			for ( String tempName : names) {
				if (getShortName(tempName).equals(shortName)
						&& (!getPackageName(tempName).startsWith("java") && !getPackageName(
								tempName).startsWith("javax"))) {
					found++;
					fullName = tempName;
					if (found > 1) {
						break;
					}
				}
			}

			// This will be the case if there are name conflicts.
			if (found > 1 || found == 0) {
				fullName = MagicClass.MAGIC_PACKAGE + "." + shortName;
			}
		}

		return fullName;
	}

	private String getPackageName(String fullName) {
		 int index = fullName.lastIndexOf(".");
		String packageName = fullName;

		if (index > -1) {
			packageName = fullName.substring(0, index);
		}

		return packageName;
	}

	private String getShortName(String fullName) {
		 int index = fullName.lastIndexOf(".");
		String shortName = fullName;

		if (index > -1) {
			shortName = fullName.substring(index + 1);
		}

		return shortName;
	}

	@Override
	protected Named getTypeFromSource(FileSource source, String name) throws SemanticException {
		 String path = new File(source.path()).getParentFile().getAbsolutePath();
		 String fullName = "";
		 
		for (Object dir : ext.getOptions().source_path) {
			File fileDir = (File) dir;
			String absPath = fileDir.getAbsolutePath();
			if (path.startsWith(fileDir.getAbsolutePath())) {
				path = path.substring(absPath.length()+1);
				path = path.replace(File.separator, ".");
				fullName = path + "." + name;
				break;
			}
		}
		if (!Options.v().classes().contains(name) && !Options.v().classes().contains(fullName)
				|| PPAEngine.v().getMode() != PPAEngine.COLLECTIVE_MODE) {
			throw new NoClassException(name);
		}
		return super.getTypeFromSource(source, name);
	}

}
